package org.apache.bcel.verifier.structurals;

import org.apache.bcel.generic.ConstantPoolGen;
import org.apache.bcel.generic.IndexedInstruction;
import org.apache.bcel.generic.LocalVariableInstruction;
import org.apache.bcel.verifier.exc.AssertionViolatedException;
import java.util.List;
import java.util.Iterator;
import org.apache.bcel.generic.CodeExceptionGen;
import java.util.Set;
import org.apache.bcel.verifier.exc.StructuralCodeConstraintException;
import java.util.ArrayList;
import org.apache.bcel.generic.ASTORE;
import java.util.HashSet;
import java.util.HashMap;
import org.apache.bcel.generic.MethodGen;
import org.apache.bcel.generic.Instruction;
import org.apache.bcel.generic.Select;
import org.apache.bcel.generic.BranchInstruction;
import org.apache.bcel.generic.GotoInstruction;
import org.apache.bcel.generic.JsrInstruction;
import org.apache.bcel.generic.ATHROW;
import org.apache.bcel.generic.ReturnInstruction;
import org.apache.bcel.generic.RET;
import org.apache.bcel.generic.InstructionHandle;
import java.util.Map;

public class Subroutines
{
    private final Map<InstructionHandle, Subroutine> subroutines;
    public final Subroutine TOPLEVEL;
    
    private static InstructionHandle[] getSuccessors(final InstructionHandle instruction) {
        final InstructionHandle[] single = { null };
        final Instruction inst = instruction.getInstruction();
        if (inst instanceof RET || inst instanceof ReturnInstruction || inst instanceof ATHROW) {
            return InstructionHandle.EMPTY_ARRAY;
        }
        if (inst instanceof JsrInstruction) {
            single[0] = instruction.getNext();
            return single;
        }
        if (inst instanceof GotoInstruction) {
            single[0] = ((GotoInstruction)inst).getTarget();
            return single;
        }
        if (!(inst instanceof BranchInstruction)) {
            single[0] = instruction.getNext();
            return single;
        }
        if (inst instanceof Select) {
            final InstructionHandle[] matchTargets = ((Select)inst).getTargets();
            final InstructionHandle[] ret = new InstructionHandle[matchTargets.length + 1];
            ret[0] = ((Select)inst).getTarget();
            System.arraycopy(matchTargets, 0, ret, 1, matchTargets.length);
            return ret;
        }
        final InstructionHandle[] pair = { instruction.getNext(), ((BranchInstruction)inst).getTarget() };
        return pair;
    }
    
    public Subroutines(final MethodGen mg) {
        this(mg, true);
    }
    
    public Subroutines(final MethodGen mg, final boolean enableJustIceCheck) {
        this.subroutines = new HashMap<InstructionHandle, Subroutine>();
        final InstructionHandle[] all = mg.getInstructionList().getInstructionHandles();
        final CodeExceptionGen[] handlers = mg.getExceptionHandlers();
        this.TOPLEVEL = new SubroutineImpl();
        final Set<InstructionHandle> subLeaders = new HashSet<InstructionHandle>();
        for (final InstructionHandle element : all) {
            final Instruction inst = element.getInstruction();
            if (inst instanceof JsrInstruction) {
                subLeaders.add(((JsrInstruction)inst).getTarget());
            }
        }
        for (final InstructionHandle astore : subLeaders) {
            final SubroutineImpl sr = new SubroutineImpl();
            sr.setLocalVariable(((ASTORE)astore.getInstruction()).getIndex());
            this.subroutines.put(astore, sr);
        }
        this.subroutines.put(all[0], this.TOPLEVEL);
        subLeaders.add(all[0]);
        for (final InstructionHandle element : all) {
            final Instruction inst = element.getInstruction();
            if (inst instanceof JsrInstruction) {
                final InstructionHandle leader = ((JsrInstruction)inst).getTarget();
                ((SubroutineImpl)this.getSubroutine(leader)).addEnteringJsrInstruction(element);
            }
        }
        final Set<InstructionHandle> instructionsAssigned = new HashSet<InstructionHandle>();
        final Map<InstructionHandle, ColourConstants> colors = new HashMap<InstructionHandle, ColourConstants>();
        final List<InstructionHandle> qList = new ArrayList<InstructionHandle>();
        for (final InstructionHandle actual : subLeaders) {
            for (final InstructionHandle element2 : all) {
                colors.put(element2, ColourConstants.WHITE);
            }
            colors.put(actual, ColourConstants.GRAY);
            qList.clear();
            qList.add(actual);
            if (actual == all[0]) {
                for (final CodeExceptionGen handler : handlers) {
                    colors.put(handler.getHandlerPC(), ColourConstants.GRAY);
                    qList.add(handler.getHandlerPC());
                }
            }
            while (!qList.isEmpty()) {
                final InstructionHandle u = qList.remove(0);
                final InstructionHandle[] successors2;
                final InstructionHandle[] successors = successors2 = getSuccessors(u);
                for (final InstructionHandle successor : successors2) {
                    if (colors.get(successor) == ColourConstants.WHITE) {
                        colors.put(successor, ColourConstants.GRAY);
                        qList.add(successor);
                    }
                }
                colors.put(u, ColourConstants.BLACK);
            }
            for (final InstructionHandle element2 : all) {
                if (colors.get(element2) == ColourConstants.BLACK) {
                    ((SubroutineImpl)((actual == all[0]) ? this.getTopLevel() : this.getSubroutine(actual))).addInstruction(element2);
                    if (instructionsAssigned.contains(element2)) {
                        throw new StructuralCodeConstraintException("Instruction '" + element2 + "' is part of more than one subroutine (or of the top level and a subroutine).");
                    }
                    instructionsAssigned.add(element2);
                }
            }
            if (actual != all[0]) {
                ((SubroutineImpl)this.getSubroutine(actual)).setLeavingRET();
            }
        }
        if (enableJustIceCheck) {
            for (final CodeExceptionGen handler2 : handlers) {
                for (InstructionHandle _protected = handler2.getStartPC(); _protected != handler2.getEndPC().getNext(); _protected = _protected.getNext()) {
                    for (final Subroutine sub : this.subroutines.values()) {
                        if (sub != this.subroutines.get(all[0]) && sub.contains(_protected)) {
                            throw new StructuralCodeConstraintException("Subroutine instruction '" + _protected + "' is protected by an exception handler, '" + handler2 + "'. This is forbidden by the JustIce verifier due to its clear definition of subroutines.");
                        }
                    }
                }
            }
        }
        this.noRecursiveCalls(this.getTopLevel(), new HashSet<Integer>());
    }
    
    public Subroutine getSubroutine(final InstructionHandle leader) {
        final Subroutine ret = this.subroutines.get(leader);
        if (ret == null) {
            throw new AssertionViolatedException("Subroutine requested for an InstructionHandle that is not a leader of a subroutine.");
        }
        if (ret == this.TOPLEVEL) {
            throw new AssertionViolatedException("TOPLEVEL special subroutine requested; use getTopLevel().");
        }
        return ret;
    }
    
    public Subroutine getTopLevel() {
        return this.TOPLEVEL;
    }
    
    private void noRecursiveCalls(final Subroutine sub, final Set<Integer> set) {
        final Subroutine[] subSubs;
        final Subroutine[] subs = subSubs = sub.subSubs();
        for (final Subroutine sub2 : subSubs) {
            final int index = ((RET)sub2.getLeavingRET().getInstruction()).getIndex();
            if (!set.add(index)) {
                final SubroutineImpl si = (SubroutineImpl)sub2;
                throw new StructuralCodeConstraintException("Subroutine with local variable '" + si.localVariable + "', JSRs '" + si.theJSRs + "', RET '" + si.theRET + "' is called by a subroutine which uses the same local variable index as itself; maybe even a recursive call? JustIce's clean definition of a subroutine forbids both.");
            }
            this.noRecursiveCalls(sub2, set);
            set.remove(index);
        }
    }
    
    public Subroutine subroutineOf(final InstructionHandle any) {
        for (final Subroutine s : this.subroutines.values()) {
            if (s.contains(any)) {
                return s;
            }
        }
        System.err.println("DEBUG: Please verify '" + any.toString(true) + "' lies in dead code.");
        return null;
    }
    
    @Override
    public String toString() {
        return "---\n" + this.subroutines + "\n---\n";
    }
    
    private enum ColourConstants
    {
        WHITE, 
        GRAY, 
        BLACK;
    }
    
    private class SubroutineImpl implements Subroutine
    {
        private static final int UNSET = -1;
        private final SubroutineImpl[] EMPTY_ARRAY;
        private int localVariable;
        private final Set<InstructionHandle> instructions;
        private final Set<InstructionHandle> theJSRs;
        private InstructionHandle theRET;
        
        public SubroutineImpl() {
            this.EMPTY_ARRAY = new SubroutineImpl[0];
            this.localVariable = -1;
            this.instructions = new HashSet<InstructionHandle>();
            this.theJSRs = new HashSet<InstructionHandle>();
        }
        
        private void _getRecursivelyAccessedLocalsIndicesHelper(final Set<Integer> set, final Subroutine[] subs) {
            for (final Subroutine sub : subs) {
                final int[] accessedLocalsIndices;
                final int[] lvs = accessedLocalsIndices = sub.getAccessedLocalsIndices();
                for (final int lv : accessedLocalsIndices) {
                    set.add(lv);
                }
                if (sub.subSubs().length != 0) {
                    this._getRecursivelyAccessedLocalsIndicesHelper(set, sub.subSubs());
                }
            }
        }
        
        public void addEnteringJsrInstruction(final InstructionHandle jsrInst) {
            if (jsrInst == null || !(jsrInst.getInstruction() instanceof JsrInstruction)) {
                throw new AssertionViolatedException("Expecting JsrInstruction InstructionHandle.");
            }
            if (this.localVariable == -1) {
                throw new AssertionViolatedException("Set the localVariable first!");
            }
            if (this.localVariable != ((ASTORE)((JsrInstruction)jsrInst.getInstruction()).getTarget().getInstruction()).getIndex()) {
                throw new AssertionViolatedException("Setting a wrong JsrInstruction.");
            }
            this.theJSRs.add(jsrInst);
        }
        
        void addInstruction(final InstructionHandle ih) {
            if (this.theRET != null) {
                throw new AssertionViolatedException("All instructions must have been added before invoking setLeavingRET().");
            }
            this.instructions.add(ih);
        }
        
        @Override
        public boolean contains(final InstructionHandle inst) {
            return this.instructions.contains(inst);
        }
        
        @Override
        public int[] getAccessedLocalsIndices() {
            final Set<Integer> acc = new HashSet<Integer>();
            if (this.theRET == null && this != Subroutines.this.getTopLevel()) {
                throw new AssertionViolatedException("This subroutine object must be built up completely before calculating accessed locals.");
            }
            for (final InstructionHandle ih : this.instructions) {
                if (ih.getInstruction() instanceof LocalVariableInstruction || ih.getInstruction() instanceof RET) {
                    final int idx = ((IndexedInstruction)ih.getInstruction()).getIndex();
                    acc.add(idx);
                    try {
                        if (!(ih.getInstruction() instanceof LocalVariableInstruction)) {
                            continue;
                        }
                        final int s = ((LocalVariableInstruction)ih.getInstruction()).getType(null).getSize();
                        if (s != 2) {
                            continue;
                        }
                        acc.add(idx + 1);
                    }
                    catch (final RuntimeException re) {
                        throw new AssertionViolatedException("BCEL did not like NULL as a ConstantPoolGen object.", re);
                    }
                }
            }
            final int[] ret = new int[acc.size()];
            int j = -1;
            for (final Integer accessedLocal : acc) {
                ++j;
                ret[j] = accessedLocal;
            }
            return ret;
        }
        
        @Override
        public InstructionHandle[] getEnteringJsrInstructions() {
            if (this == Subroutines.this.getTopLevel()) {
                throw new AssertionViolatedException("getLeavingRET() called on top level pseudo-subroutine.");
            }
            return this.theJSRs.toArray(InstructionHandle.EMPTY_ARRAY);
        }
        
        @Override
        public InstructionHandle[] getInstructions() {
            return this.instructions.toArray(InstructionHandle.EMPTY_ARRAY);
        }
        
        @Override
        public InstructionHandle getLeavingRET() {
            if (this == Subroutines.this.getTopLevel()) {
                throw new AssertionViolatedException("getLeavingRET() called on top level pseudo-subroutine.");
            }
            return this.theRET;
        }
        
        @Override
        public int[] getRecursivelyAccessedLocalsIndices() {
            final Set<Integer> s = new HashSet<Integer>();
            final int[] accessedLocalsIndices;
            final int[] lvs = accessedLocalsIndices = this.getAccessedLocalsIndices();
            for (final int lv : accessedLocalsIndices) {
                s.add(lv);
            }
            this._getRecursivelyAccessedLocalsIndicesHelper(s, this.subSubs());
            final int[] ret = new int[s.size()];
            int j = -1;
            for (final Integer index : s) {
                ++j;
                ret[j] = index;
            }
            return ret;
        }
        
        void setLeavingRET() {
            if (this.localVariable == -1) {
                throw new AssertionViolatedException("setLeavingRET() called for top-level 'subroutine' or forgot to set local variable first.");
            }
            InstructionHandle ret = null;
            for (final InstructionHandle actual : this.instructions) {
                if (actual.getInstruction() instanceof RET) {
                    if (ret != null) {
                        throw new StructuralCodeConstraintException("Subroutine with more then one RET detected: '" + ret + "' and '" + actual + "'.");
                    }
                    ret = actual;
                }
            }
            if (ret == null) {
                throw new StructuralCodeConstraintException("Subroutine without a RET detected.");
            }
            if (((RET)ret.getInstruction()).getIndex() != this.localVariable) {
                throw new StructuralCodeConstraintException("Subroutine uses '" + ret + "' which does not match the correct local variable '" + this.localVariable + "'.");
            }
            this.theRET = ret;
        }
        
        void setLocalVariable(final int i) {
            if (this.localVariable != -1) {
                throw new AssertionViolatedException("localVariable set twice.");
            }
            this.localVariable = i;
        }
        
        @Override
        public Subroutine[] subSubs() {
            final Set<Subroutine> h = new HashSet<Subroutine>();
            for (final InstructionHandle ih : this.instructions) {
                final Instruction inst = ih.getInstruction();
                if (inst instanceof JsrInstruction) {
                    final InstructionHandle targ = ((JsrInstruction)inst).getTarget();
                    h.add(Subroutines.this.getSubroutine(targ));
                }
            }
            return h.toArray(this.EMPTY_ARRAY);
        }
        
        @Override
        public String toString() {
            final StringBuilder ret = new StringBuilder();
            ret.append("Subroutine: Local variable is '").append(this.localVariable);
            ret.append("', JSRs are '").append(this.theJSRs);
            ret.append("', RET is '").append(this.theRET);
            ret.append("', Instructions: '").append(this.instructions).append("'.");
            ret.append(" Accessed local variable slots: '");
            final int[] accessedLocalsIndices;
            int[] alv = accessedLocalsIndices = this.getAccessedLocalsIndices();
            for (final int element : accessedLocalsIndices) {
                ret.append(element);
                ret.append(" ");
            }
            ret.append("'.");
            ret.append(" Recursively (via subsub...routines) accessed local variable slots: '");
            final int[] recursivelyAccessedLocalsIndices;
            alv = (recursivelyAccessedLocalsIndices = this.getRecursivelyAccessedLocalsIndices());
            for (final int element : recursivelyAccessedLocalsIndices) {
                ret.append(element);
                ret.append(" ");
            }
            ret.append("'.");
            return ret.toString();
        }
    }
}
